import zstandard


_zdec = zstandard.ZstdDecompressor()


def decompress(b: bytes, uncompressed_size: int) -> bytes:
    return _zdec.decompress(b, uncompressed_size)


def maybe_decompress(b: bytes) -> bytes:
    if b[0] == 2:  # compressed v2 format
        return _zdec.decompress(
            b[1:], max_output_size=2147483648
        )  # https://stackoverflow.com/questions/69270987/how-to-resolve-the-error-related-to-frame-used-in-zstandard-which-requires-too-m
    return b



import hashlib
import logging
import os
from pathlib import Path

import pyrrhic.cli.state
from pyrrhic.cli.restore import restore

import pytest

from .params import params

RESTORE_FILES = [
    (Path("usr/share/cracklib/cracklib.magic"), "c4b2b3034acf5b35b60a8de27c7ac33f54c6b4ea"),
    (Path("usr/share/cracklib/cracklib-double"), "5fc0bc9ea625e08be635226518d47d3070accc5b"),
    (Path("usr/share/cracklib/cracklib-small"), "5f97502ab12eac2e3aa869d00a13af41a7f585e6"),
    (Path("usr/share/cracklib/pw_dict.hwm"), "5dfc5fa9b8fec7eff807ede3561f4b2cdca17277"),
    (Path("usr/share/cracklib/pw_dict.pwd"), "20973efecb1e0239800a6e2437fa3390d3fd415c"),
    (Path("usr/share/cracklib/pw_dict.pwi"), "398936961dff2e5f710723c36e67edca92943284"),
]


@pytest.mark.parametrize("repository,snapshot_cracklib", [(p["repo"], p["snapshot_cracklib"]) for p in params])
def test_restore(capfd, tmp_path, caplog, repository, snapshot_cracklib):
    pyrrhic.cli.state.repository = repository
    restore(snapshot_cracklib, target=tmp_path)
    for snapshot_path, sha1sum in RESTORE_FILES:
        assert hashlib.sha1((tmp_path / snapshot_path).read_bytes()).hexdigest() == sha1sum
    # Resume backup
    resume_from = os.stat(tmp_path / "usr/share/cracklib/cracklib-double").st_size - 10
    caplog.set_level(logging.DEBUG)
    logging.debug(f"Should resume from {resume_from}")
    with open(tmp_path / "usr/share/cracklib/cracklib-double", "a") as f:
        f.truncate(resume_from)
        logging.debug(f"truncated {tmp_path / 'usr/share/cracklib/cracklib-double'}")

    restore(snapshot_cracklib, target=tmp_path, resume=True)
    assert len([record.msg for record in caplog.records if "Resuming from" in record.msg]) == 1

    for snapshot_path, sha1sum in RESTORE_FILES:
        assert hashlib.sha1((tmp_path / snapshot_path).read_bytes()).hexdigest() == sha1sum

    symlink_path = Path(tmp_path / "usr/share/cracklib/cracklib-dummy")
    assert symlink_path.is_symlink()
    assert symlink_path.resolve().name == "cracklib-small"



import operator
import stat

import pyrrhic.cli.state as state
from pyrrhic.repo.tree import Node, walk_breadth_first

from rich import print
from rich.table import Table


import typer

_MODE_STAT = {
    "dir": stat.S_IFDIR,
    "file": stat.S_IFREG,
    "symlink": stat.S_IFLNK,
    "dev": stat.S_IFBLK,
    "chardev": stat.S_IFCHR,
    "socket": stat.S_IFSOCK,
    "fifo": stat.S_IFIFO,
}


def _format_name(node: Node) -> str:
    if node.type == "symlink":
        return f"{node.name} -> {node.linktarget}"
    return node.name


def _print_long(node: Node, path: str, table: Table) -> None:
    name = _format_name(node)
    mtime_str = f"{node.mtime:%Y-%m-%d %H:%M:%S}"  # rich compatible datetime str
    mode = stat.filemode(stat.S_IMODE(node.mode) | _MODE_STAT.get(node.type, 0))
    table.add_row(f"{mode}", f"{node.uid}", f"{node.gid}", f"{node.size}", f"{mtime_str}", f"{path}/{name}")


def ls(snapshot_prefix: str, long: bool = typer.Option(False, "--long", "-l", help="Use a long listing (Unix ls -l)")):
    "List files in a snapshot"
    state.repository.get_snapshot(snapshot_prefix)
    if snapshot_prefix == "latest":
        snapshots = iter(sorted(state.repository.get_snapshot(), key=operator.attrgetter("time"), reverse=True)[:1])
    else:
        snapshots = state.repository.get_snapshot(snapshot_prefix)
    snapshot = next(snapshots, None)
    if not snapshot:
        raise ValueError(f"Index: {snapshot_prefix} not found")
    if next(snapshots, None):
        raise ValueError(f"Prefix {snapshot_prefix} matches multiple snapshots")
    if long:
        table = Table("mode", "user", "group", "size", "date", "filename", highlight=True)
        for pleaf in walk_breadth_first(state.repository, snapshot.tree):
            _print_long(pleaf.node, pleaf.path, table)
        print(table)
    else:
        for pleaf in walk_breadth_first(state.repository, snapshot.tree):
            name = _format_name(pleaf.node)
            print(f"{pleaf.path}/{name}")



import hashlib
import json
from ast import literal_eval
from pathlib import Path

import pyrrhic.cli.state
from pyrrhic.cli.cat import config, index, masterkey, pack, snapshot
from pyrrhic.cli.ls import ls

import pytest

from .params import params


@pytest.mark.parametrize("repository,expected", [(p["repo"], p["masterkey"]) for p in params])
def test_cat_masterkey(capfd, repository, expected):
    pyrrhic.cli.state.repository = repository
    masterkey()
    assert literal_eval(capfd.readouterr().out) == expected


@pytest.mark.parametrize("repository,config_dict", [(p["repo"], p["config"]) for p in params])
def test_cat_config(capfd, repository, config_dict):
    pyrrhic.cli.state.repository = repository
    config()
    assert literal_eval(capfd.readouterr().out) == config_dict


@pytest.mark.parametrize("repository,snapshot_id,snapshot_json", [(p["repo"], p["snapshot_id"], p["snapshot_json"]) for p in params])
def test_cat_snapshot(capfd, repository, snapshot_id, snapshot_json):
    pyrrhic.cli.state.repository = repository
    snapshot(snapshot_id)
    out = capfd.readouterr().out
    assert json.loads(out) == snapshot_json


@pytest.mark.parametrize("repository,index_id,index_substr", [(p["repo"], p["index_id"], p["index_substr"]) for p in params])
def test_cat_index(capfd, repository, index_id, index_substr):
    pyrrhic.cli.state.repository = repository
    index(index_id)
    out = capfd.readouterr().out
    assert "PackRef" in out
    assert index_substr in out


@pytest.mark.parametrize("repository,pack_id", [(p["repo"], p["pack_id"]) for p in params])
def test_cat_pack(capfdbinary, repository, pack_id):
    pyrrhic.cli.state.repository = repository
    sha = pack_id
    pack(sha, False)
    assert hashlib.sha256(capfdbinary.readouterr().out).hexdigest() == sha


@pytest.mark.parametrize("repository,pack_id,pack_blobs", [(p["repo"], p["pack_id"], p["pack_blobs"]) for p in params])
def test_cat_pack_header(capfd, repository, pack_id, pack_blobs):
    pyrrhic.cli.state.repository = repository
    sha = pack_id
    pack(sha, True)
    assert literal_eval(capfd.readouterr().out) == pack_blobs


@pytest.mark.parametrize("repository", [(p["repo"]) for p in params])
def test_ls(capfd, repository):
    "Returns a list of paths"
    pyrrhic.cli.state.repository = repository
    ls("latest", False)
    lines = capfd.readouterr().out.splitlines()
    for s in lines:
        path = Path(s)
        assert path.is_absolute()
    with pytest.raises(ValueError, match="Index: invalid not found"):
        ls("invalid", False)
    with pytest.raises(ValueError, match="Prefix  matches multiple snapshots"):
        ls("", False)



import hashlib
import io
from collections import OrderedDict
from copy import copy
from dataclasses import dataclass
from datetime import datetime
from logging import info
from pathlib import Path
from typing import Iterator, Optional

import msgspec

from pyrrhic.crypto.keys import decrypt_mac
from pyrrhic.repo.repository import Repository
from pyrrhic.util import decompress


class Node(msgspec.Struct):
    name: str
    type: str
    mode: int
    mtime: datetime
    atime: datetime
    ctime: datetime
    uid: int
    gid: int
    size: int = 0
    user: str
    inode: int
    content: Optional[list[str]] = None
    linktarget: Optional[str] = None
    subtree: Optional[str] = None


class Tree(msgspec.Struct):
    nodes: list[Node]


class ReaderCache:
    def __init__(self, capacity: int):
        self.cache: OrderedDict[Path, io.BufferedReader] = OrderedDict()
        self.misses = 0  # stats
        self.requests = 0
        self.capacity = capacity

    def get(self, key: Path) -> io.BufferedReader:
        self.requests += 1
        if key not in self.cache:
            self.cache[key] = open(key, "rb")
            self.misses += 1
            if len(self.cache) > self.capacity:
                self.cache.popitem(last=False)
        self.cache.move_to_end(key)
        return self.cache[key]

    def flush(self):
        for item in self.cache.values():
            item.close()
        self.cache = OrderedDict()
        misses = self.misses / self.requests
        info(f"Cache misses: {misses:.2%}")


def get_node_blob(repo: Repository, rcache: ReaderCache, blob_id: str) -> bytes:
    index = repo.get_index()
    packref = index.get_packref(blob_id)
    blob = packref.blob
    f = rcache.get(repo.repository / "data" / packref.id[:2] / packref.id)
    f.seek(blob.offset)
    buffer = f.read(blob.length)
    plaintext = decrypt_mac(repo.masterkey, buffer)
    if uncompressed_length := blob.uncompressed_length:
        plaintext = decompress(plaintext, uncompressed_length)
    if hashlib.sha256(plaintext).hexdigest() != blob.id:
        raise ValueError(f"Invalid hash for blob {blob.id}")
    return plaintext


# FIXME: Use only one msgspec object
def get_tree(repo: Repository, rcache: ReaderCache, tree_id: str) -> Tree:
    plaintext_blob = get_node_blob(repo, rcache, tree_id)
    return msgspec.json.decode(plaintext_blob, type=Tree)


@dataclass(frozen=True)
class PathNode:
    "Represent a Node located at path"
    path: str  # prefix path
    node: Node


def walk_breadth_first(repository: Repository, tree_id: str, rcache=ReaderCache(64)) -> Iterator[PathNode]:
    tree = get_tree(repository, rcache, tree_id)
    pathnodes = [PathNode(f"/{node.name}", node) for node in tree.nodes]
    while pathnodes:
        pleafes = [pnode for pnode in pathnodes if not pnode.node.subtree]
        pathnodes = [pnode for pnode in pathnodes if pnode.node.subtree]  # FIXME: traverses 2 times
        for pleaf in pleafes:
            yield pleaf
        if pathnodes:
            pnode = pathnodes.pop()
            node_without_subdir = copy(pnode.node)
            node_without_subdir.subtree = None
            tree = get_tree(repository, rcache, pnode.node.subtree or "0123")  # just make mypy happy?
            pathnodes = [PathNode(pnode.path, node_without_subdir), *[PathNode(f"{pnode.path}/{node.name}", node) for node in tree.nodes], *pathnodes]
    rcache.flush()



"""
This module contains the crypto primitives used by restic.
"""

import json
from base64 import b64decode, b64encode
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import poly1305
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives.kdf.scrypt import Scrypt

import msgspec

# mask for key, (cf. http://cr.yp.to/mac/poly1305-20050329.pdf)
_POLY1305KEYMASK = b"\xff\xff\xff\x0f\xfc\xff\xff\x0f\xfc\xff\xff\x0f\xfc\xff\xff\x0f"


class WrappedKey(msgspec.Struct):
    """Class that contain all data that is needed to derive the
    repository's master encryption and message authentication keys
    from a user's password."""

    hostname: str
    username: str
    kdf: str
    N: int
    r: int
    p: int
    created: datetime
    data: bytes
    salt: bytes


@dataclass(frozen=True)
class Mac:
    """Class that holdes the Poly1305-AES parameters"""

    k: bytes
    r: bytes


@dataclass(frozen=True)
class MasterKey:
    """Class that holds encryption and message authentication keys for a
    repository."""

    mac: Mac
    encryption: bytes

    def restic_json(self):
        """Return restic representation of Masterkey"""
        return {
            "mac": {
                "r": b64encode(self.mac.r).decode(),
                "k": b64encode(self.mac.k).decode(),
            },
            "encrypt": b64encode(self.encryption).decode(),
        }

    def __hash__(self):
        return hash(self.encryption + self.mac.k + self.mac.r)


def load_key(key_path: Path) -> WrappedKey:
    return msgspec.json.decode(key_path.read_bytes(), type=WrappedKey)


def _poly1305_validate(nonce: bytes, k: bytes, r: bytes, message: bytes, mac: bytes) -> bool:
    r = bytes([(r & m) for r, m in zip(r, _POLY1305KEYMASK)])
    cipher = Cipher(algorithms.AES(k), modes.ECB())
    encryptor = cipher.encryptor()
    aes_ciphertext = encryptor.update(nonce) + encryptor.finalize()
    poly1305_key = r + aes_ciphertext
    p = poly1305.Poly1305(poly1305_key)
    p.update(message)
    return p.finalize() == mac


def _decrypt(aes_key: bytes, nonce: bytes, ciphertext: bytes) -> bytes:
    cipher = Cipher(algorithms.AES(aes_key), modes.CTR(nonce))
    decryptor = cipher.decryptor()
    return decryptor.update(ciphertext) + decryptor.finalize()


def decrypt_mac(key: MasterKey, restic_blob: bytes) -> bytes:
    """Decrypt 'IV || CIPHERTEXT || MAC' bytes, validate mac and
    return plaintext bytes"""
    nonce = restic_blob[:16]
    mac = restic_blob[-16:]
    ciphertext = restic_blob[16:-16]
    if not _poly1305_validate(nonce, key.mac.k, key.mac.r, ciphertext, mac):
        raise ValueError("ciphertext verification failed")
    return _decrypt(key.encryption, nonce, ciphertext)


def get_masterkey(path: Path, password: bytes) -> MasterKey:
    wrapped_key = load_key(path)
    kdf = Scrypt(
        wrapped_key.salt,
        64,
        wrapped_key.N,
        wrapped_key.r,
        wrapped_key.p,
        backend=default_backend,
    )
    derived_key = kdf.derive(password)
    poly_k = derived_key[32:48]
    poly_r = derived_key[48:]
    nonce = wrapped_key.data[0:16]
    message = wrapped_key.data[16:-16]
    mac = wrapped_key.data[-16:]
    if not _poly1305_validate(nonce, poly_k, poly_r, message, mac):
        raise ValueError("Invalid Password")
    j = json.loads(_decrypt(derived_key[:32], nonce, message))
    encryption = b64decode(j["encrypt"])
    r = b64decode(j["mac"]["r"])
    k = b64decode(j["mac"]["k"])
    return MasterKey(Mac(k=k, r=r), encryption)



from pathlib import Path

from pyrrhic.repo.repository import Repository, get_masterkey

params = [
    {
        "repo": Repository(
            Path("restic_test_repositories/restic_test_repository"), get_masterkey(Path("restic_test_repositories/restic_test_repository"), "password")
        ),
        "config": {"chunker_polynomial": "3833148ec41f8d", "id": "2ec6792a9c75a017ec4665a5e7733f45f8bffb67c7d0f3c9ec6cc96e58c6386b", "version": 1},
        "masterkey": {"mac": {"r": "hQYBDSD/JwpU8XMDIJmAAg==", "k": "aSbwRFL8rIOOxL4W+mAW1w=="}, "encrypt": "Te0IPiu0wvEtr2+J59McgTrjCp/ynVxC/mmM9mX/t+E="},
        "snapshot_id": "dd62b535d10bd8f24440cc300a868d6bf2f472859f1218883b0a6faca364c10c",
        "snapshot_json": {
            "time": "2022-07-19T19:52:28.692252Z",
            "tree": "a5dbcc77f63f5dd4f4c67c988aba4a19817aaa9d6c34a6021236a5d40ce653e1",
            "paths": ["/home/juergen/shared/python/pyrrhic/test"],
            "hostname": "shaun",
            "username": "juergen",
            "id": "dd62b535d10bd8f24440cc300a868d6bf2f472859f1218883b0a6faca364c10c",
            "uid": 1000,
            "gid": 1000,
            "excludes": None,
            "tags": None,
        },
        "snapshot_cracklib": "d2bbc914",
        "index_id": "bd10c8e85d2cdc3267faae1748cf7a334385d021766059580976882556097c0d",
        "index_substr": """'511d1c632ccad135d5407157154eccc17fcfaf501ad252b231c7ce41175473b9': PackRef(
        id='46771395523ccd6dda16694f0ce775f9508a4c3e4527c385f55d8efafa36807f'""",
        "pack_id": "4b24375f07a164e995d06303bfc26f79f94127e4e5c6e1c476495bbee0af7ccc",
        "pack_blobs": [
            {"id": "82deec86c5611cb5ae02b967e49d7aeaca50a732432bd1a5923787bd5d0fbf80", "length": 1643, "offset": 0},
            {"id": "e20d6400fbd4e602c79c8bab98a88726865b168838cc8107d560da10f19b2ff8", "length": 1467, "offset": 1643},
            {"id": "a5dbcc77f63f5dd4f4c67c988aba4a19817aaa9d6c34a6021236a5d40ce653e1", "length": 413, "offset": 3110},
        ],
    },
    {
        "repo": Repository(
            Path("restic_test_repositories/restic_v2_test_repository"), get_masterkey(Path("restic_test_repositories/restic_v2_test_repository"), "password")
        ),
        "config": {"version": 2, "id": "2c5d1276aa6f58c06883c2a7a2aef35c86ec4a0b57a6ded80e08de3fedc108f8", "chunker_polynomial": "232a535caedb65"},
        "masterkey": {"mac": {"r": "2iQMCQDOPw1UtW0N1C4RAA==", "k": "zpL4Veh3FhT7grSgm0BmCQ=="}, "encrypt": "193cMViwx7TJFtykWOZzEKojzp0fjx4VvaiJ6yMgZ3E="},
        "snapshot_id": "4d9c1f",
        "snapshot_json": {
            "time": "2022-10-18T04:53:22.965183Z",
            "tree": "9178aa841c9fd277478925c7864da28e4500ce236ad89a0ff9a261b1dabbda42",
            "paths": ["/usr/share/cracklib"],
            "hostname": "shaun",
            "username": "juergen",
            "id": "4d9c1f12d69290f6c3cc8984ef08b8e17b233fddc1bb1b0b2f9a669620f17a59",
            "uid": 1000,
            "gid": 1000,
            "excludes": None,
            "tags": None,
        },
        "snapshot_cracklib": "78102f",
        "index_id": "2cd9a598873e58bcb067b1d3f6db488047a2bc168da7550b8340c319f0693d8a",
        "index_substr": """'8696358373607539e46c42a13955e343cd1d66a45b102c3ccf41b5e38d4b1db1': PackRef(
        id='7c42983e65e74be3911ab1c9177e10128bce39d0c29033a2db9de4a6810d0711'""",
        "pack_id": "ed505dcfdd8b9c7cb912be91bc19b711279253e37358f70af65466c92816febb",
        "pack_blobs": [
            {"id": "8072950e2d305ebcfd94912557e06b6a9e1a7ee487470e4d89efd34e749a2505", "offset": 0, "length": 621, "length_uncompressed": 2197},
            {"id": "25fec2e636c582064ea0927886a93acca579ab1436b01311f18c10776aa70fa4", "offset": 621, "length": 256, "length_uncompressed": 373},
            {"id": "b9c28d59a52a1c5406637280aff6f63195fc9dac603d3e070fbd884aaddb6b5f", "offset": 877, "length": 255, "length_uncompressed": 370},
            {"id": "9178aa841c9fd277478925c7864da28e4500ce236ad89a0ff9a261b1dabbda42", "offset": 1132, "length": 253, "length_uncompressed": 368},
        ],
    },
]



import operator
import os
import stat
from logging import debug, info, warn
from pathlib import Path

import pyrrhic.cli.state as state
from pyrrhic.cli.util import catch_exception
from pyrrhic.repo.tree import ReaderCache, get_node_blob, walk_breadth_first

from rich.progress import track

import typer


def _restore(tree_id: str, target: Path, resume=False):
    isroot = os.geteuid() == 0
    rcache = ReaderCache(64)
    index = state.repository.get_index()
    for pnode in walk_breadth_first(state.repository, tree_id, rcache):
        node = pnode.node
        abs_path = target / Path(pnode.path).relative_to("/")
        mode = stat.S_IMODE(node.mode)
        match node.type:
            case "file":
                # FIXME: Create temporary unique file and do atomic rename
                if node.content:  # possible empty file
                    resume_from = 0
                    blobs = [index.get_packref(content).blob for content in node.content]
                    if abs_path.is_file():
                        if not resume:
                            raise FileExistsError(f"File exists: {abs_path}")
                        resume_from = os.stat(abs_path).st_size
                        if resume_from == node.size:
                            debug(f"Already restored {pnode.path}")
                            continue
                        if resume_from > node.size:
                            raise ValueError(f"Existing file {abs_path} is larger")
                    with open(abs_path, "ab") as f:
                        current_pos = 0
                        for i, blob in enumerate(track(blobs, pnode.path)):
                            debug(f"Processing Blob {i} of {abs_path}")
                            blength = blob.uncompressed_length or (blob.length - 32)
                            if current_pos < resume_from:
                                if resume_from < (current_pos + blength):
                                    f.truncate(current_pos)
                                    info(f"Resuming from {current_pos}")
                                else:
                                    debug(f"{abs_path} Ignoring pos {current_pos} in resumed file")
                                    current_pos += blength
                                    continue
                            bs = get_node_blob(state.repository, rcache, blob.id)
                            current_pos += len(bs)
                            f.write(bs)

                    abs_path.chmod(mode)
                    if isroot:  # FIXME: chgrp to groups this user is member of
                        os.chown(abs_path, node.uid, node.gid)

            case "dir":
                debug(f"Creating directory {abs_path}")
                try:
                    abs_path.mkdir(mode)
                except FileExistsError as e:
                    if not resume:
                        raise e
            case "symlink":
                if abs_path.is_symlink():
                    if not resume:
                        raise FileExistsError(f"Symlink exists: {abs_path}")
                    debug(f"Symlink exists: {abs_path}")
                elif node.linktarget:
                    debug(f"Creating symlink {abs_path} -> {node.linktarget}")
                    os.symlink(node.linktarget, abs_path)
                    if isroot:  # FIXME: chgrp to groups this user is member of
                        os.chown(abs_path, node.uid, node.gid, follow_symlinks=False)
                else:
                    raise ValueError(f"Symlink target of {abs_path} does not exist")
            case _:
                warn(f"{node.name}: {node.type} not implemented")


@catch_exception(OSError, exit_code=2)
def restore(
    snapshot_prefix: str,
    target: Path,
    resume: bool = typer.Option(False, "--resume", "-r", help="resume exiting files in target"),
    help="Restore data from a snapshot",
):
    state.repository.get_snapshot(snapshot_prefix)
    if snapshot_prefix == "latest":  # FIXME: Duplicated code (ls command)
        snapshots = iter(sorted(state.repository.get_snapshot(), key=operator.attrgetter("time"), reverse=True)[:1])
    else:
        snapshots = state.repository.get_snapshot(snapshot_prefix)
    snapshot = next(snapshots, None)
    if not snapshot:
        raise ValueError(f"Index: {snapshot_prefix} not found")
    if next(snapshots, None):
        raise ValueError(f"Prefix {snapshot_prefix} matches multiple snapshots")
    _restore(snapshot.tree, target, resume)



from pyrrhic.repo.index import Index
from pyrrhic.repo.pack import Pack

import pytest

from .params import params


@pytest.mark.parametrize("repo", [(p["repo"]) for p in params])
def test_index_matches_packs(repo):
    index = repo.get_index()
    assert type(index) == Index
    for blob_id, packref in index.index.items():
        p = Pack(repo.repository, repo.masterkey, packref.id)
        pack_blobs = p.get_blob_index()
        assert (matching_blob := next((blob for blob in pack_blobs if blob.hash.hex() == blob_id)))
        assert matching_blob.offset == packref.blob.offset
        assert matching_blob.length == packref.blob.length
        assert matching_blob.uncompressed_length == packref.blob.uncompressed_length
